//go:build !minimal

package lib

import (
	"crypto/tls"
	"fmt"
	"net"
	"net/http"
	"net/url"
	"sync"

	"github.com/gorilla/websocket"
)

type IRCWebSocket struct {
	writeMutex sync.Mutex
	closeOnce  sync.Once
	websocket  *websocket.Conn
}

func NewIRCWebSocket(wsUrl, origin string, tlsConfig *tls.Config) (IRCConnection, error) {
	var headers http.Header
	if origin != "" {
		headers = make(http.Header)
		u, err := url.Parse(origin)
		if err != nil {
			return nil, err
		}
		if u.Scheme == "" {
			u.Scheme = "https"
		}
		headers.Set("Origin", u.String())
	}

	dialer := websocket.Dialer{
		Subprotocols:    []string{"text.ircv3.net", "binary.ircv3.net"},
		TLSClientConfig: tlsConfig,
	}
	ws, resp, err := dialer.Dial(wsUrl, headers)
	if err != nil {
		explanation := "no HTTP response"
		if resp != nil {
			explanation = fmt.Sprintf("HTTP status code %d", resp.StatusCode)
		}
		return nil, fmt.Errorf("%w (%s)", err, explanation)
	}
	return &IRCWebSocket{
		websocket: ws,
	}, nil
}

func (w *IRCWebSocket) GetLine() (string, error) {
	_, lineBytes, err := w.websocket.ReadMessage()
	return string(lineBytes), err
}

func (w *IRCWebSocket) SendLine(line string) error {
	messageType := websocket.TextMessage
	if w.websocket.Subprotocol() == "binary.ircv3.net" {
		messageType = websocket.BinaryMessage
	}
	w.writeMutex.Lock()
	defer w.writeMutex.Unlock()
	return w.websocket.WriteMessage(messageType, []byte(line))
}

func (w *IRCWebSocket) Disconnect() {
	w.closeOnce.Do(w.realDisconnect)
}

func (w *IRCWebSocket) realDisconnect() {
	w.websocket.Close()
}

func (w *IRCWebSocket) RemoteAddr() net.Addr {
	return w.websocket.RemoteAddr()
}
